"""Device watchdog"""
import asyncio
import contextlib
from datetime import datetime
import logging

from shifu_driver.schema import ConnectionState


class Watchdog:
    def __init__(self, send_heartbeat, connection_status, watchdog_config, event_loop):
        self.logger = logging.getLogger(__name__)

        # Set user-defined heartbeat sending function
        self.send_heartbeat = send_heartbeat

        # Set device connection state, task list, config refs
        self.connection = connection_status
        self.config = watchdog_config
        self.watchdog_tasks = set()
        self.event_loop = event_loop

        # Set heartbeat sending rate
        self.send_delay = self.config.heartbeat_send_delay
        self.max_delay = 30  # Max sending delay, in seconds

        # Watchdog flags
        self.watchdog_active = False
        self.rate_limit_active = False

    def start(self):
        asyncio.run_coroutine_threadsafe(self._start(), self.event_loop)

    async def _start(self):
        self.watchdog_active = True
        self.device_alive_flag = asyncio.Event()
        self.logger.debug(f"{self.config.reference_name} watchdog starting...")

        # Create watchdog tasks
        watchdog_task = asyncio.create_task(self.run_watchdog(), name='run_watchdog')
        timeout_task = asyncio.create_task(self.check_timeout(), name='check_timeout')

        # Add watchdog tasks to set of device tasks
        self.watchdog_tasks.add(watchdog_task)
        self.watchdog_tasks.add(timeout_task)

    def stop(self):
        asyncio.run_coroutine_threadsafe(self._stop(), self.event_loop)

    async def _stop(self):
        self.watchdog_active = False
        self.connection.heartbeat_failure_count = 0

        for task in self.watchdog_tasks:
            task.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await task

    async def run_watchdog(self):
        self.logger.debug(f"{self.config.reference_name} watchdog started")

        while self.watchdog_active:
            await asyncio.sleep(self.get_send_delay())
            self.connection.last_heartbeat_sent = datetime.now()
            self.connection.heartbeat_packets_sent += 1
            # Optionally log heartbeat sent
            if self.config.heartbeat_logging:
                self.logger.debug("Send Heartbeat called")
            # before_heartbeat = time.time()
            self.send_heartbeat(self.config.heartbeat_timeout)
            # after_heartbeat = time.time()

        self.logger.debug(f"{self.config.reference_name} watchdog exiting")

    async def check_timeout(self):
        while self.watchdog_active:
            await asyncio.sleep(self.config.heartbeat_timeout)

            # Check if device heartbeat has timed out
            if (
                self.connection.last_heartbeat_sent - self.connection.last_heartbeat_received
            ).total_seconds() > self.config.heartbeat_timeout:
                self.logger.warning(f"Heartbeat timeout on {self.config.reference_name}")
                self.connection.heartbeat_failure_count += 1

            # Check if device has exceeded heartbeat failure limit
            if self.connection.heartbeat_failure_count >= self.config.heartbeat_failure_limit:
                self.logger.error(f"Heartbeat failure on {self.config.reference_name}")
                self.rate_limit_active = True
                self.connection.state = ConnectionState.WATCHDOG_FAILED

                # Wait for device to reconnect before continuing to check timeout
                self.device_alive_flag.clear()
                await self.device_alive_flag.wait()
                self.logger.info(f"{self.config.reference_name} reconnected")

    def feed_watchdog(self):
        self.connection.last_heartbeat_received = datetime.now()
        self.connection.heartbeat_packets_received += 1
        self.connection.heartbeat_failure_count = 0
        self.rate_limit_active = False

        # Signal that device is alive
        if not self.device_alive_flag.is_set():
            asyncio.run_coroutine_threadsafe(self.set_device_alive_flag(), self.event_loop)

        # Optionally log heartbeat received
        if self.config.heartbeat_logging:
            self.logger.debug("Heartbeat received")

    def get_send_delay(self):
        if not self.rate_limit_active:
            self.send_delay = self.config.heartbeat_send_delay
        elif self.send_delay < self.max_delay:
            self.send_delay = min(self.send_delay * 2, self.max_delay)
        else:
            self.send_delay = self.max_delay
        return self.send_delay

    async def set_device_alive_flag(self):
        self.device_alive_flag.set()

        # Reset to valid connection state if in watchdog failed state
        if self.connection.state == ConnectionState.WATCHDOG_FAILED:
            self.connection.state = ConnectionState.CONNECTED



"""Base implementation for device drivers"""
import abc
import asyncio
import json
import logging

from flask import Flask, current_app, make_response, request, jsonify

from shifu_driver import schema, watchdog, socket_backend, driver_bp


def route(route, endpoint=None, methods=['POST']):
    def wrapped_f(func):
        func.__route__ = route
        func.__endpoint__ = endpoint
        func.__methods__ = methods
        return func

    return wrapped_f


def is_route(func):
    return hasattr(func, '__route__') and hasattr(func, '__methods__')


def view_func_wrapper(func):
    def view_func():
        json_dict = request.get_json(silent=True) or {}
        return func(**json_dict)

    return view_func


class ShifuDriver(abc.ABC, Flask):
    def __init__(self):
        super().__init__(__name__)

        # Start logging
        self.logger = logging.getLogger(__name__)

        # Set asyncio event loop
        self.event_loop = asyncio.get_event_loop()

        # Create device state attributes
        self.device_status = schema.DeviceStatus()
        self.connection_status = schema.ConnectionStatus()

        # # Override self.connect in child classes in needed
        self.initialized = False

        driver_bp.before_request_funcs.setdefault(None, []).append(self.check_device_initialization)
        for attr in dir(self):
            field = getattr(self, attr)
            if callable(field) and is_route(field):
                driver_bp.add_url_rule(
                    field.__route__,
                    field.__endpoint__ or '-'.join(field.__qualname__.split('.')),
                    view_func_wrapper(field),
                    methods=field.__methods__,
                )

        self.register_blueprint(driver_bp)

    def __del__(self):
        if self.initialized:
            self.disconnect()
            self.watchdog.stop()

    def run(self, host=None, port=None, debug=None, load_dotenv=True, **options):
        super().run(host, port, debug, load_dotenv, **options)

    def check_device_initialization(self):
        if not self.initialized and (
            request.endpoint not in ['driver.ShifuDriver-initialize_device', 'driver.ShifuDriver-_endpoints']
        ):
            return "Device not initialized"

    @route('/initialize', methods=['POST'])
    def initialize_device(self, **kwargs):
        self.device_config = schema.Config.from_json(kwargs)
        self.com_backend = self.connect()
        self.watchdog = watchdog.Watchdog(
            self.send_heartbeat, self.connection_status, self.device_config, self.event_loop
        )
        self.watchdog.start()
        self.initialized = True
        return make_response("Initialization Success!")

    @route('/endpoints', methods=['GET'])
    def _endpoints(self):
        url_map = current_app.url_map

        device_endpoints = []
        for rule in url_map._rules:
            device_endpoint = {'rule': rule.rule, 'endpoint': rule.endpoint, 'methods': list(rule.methods)}
            device_endpoint['view_func'] = current_app.view_functions[device_endpoint['endpoint']].__qualname__
            device_endpoints.append(device_endpoint)

        return jsonify(device_endpoints)

    # TODO: refactor according to current config,
    # TODO: connection_info should be from a different source if com_backend is not used
    def to_dict(self):
        data = {'connection_info': self.connection_status.to_dict(), 'device_info': self.device_status.to_dict()}
        return data

    def __str__(self):
        return json.dumps(self.to_dict(), indent=4)

    def __repr__(self):
        return f"<{type(self).__name__} ({self.__class__!r})>"

    # ############ Standard device functions ############ #

    # Externally accessible connect method
    def connect(self):
        com_backend = socket_backend.SocketBackend(
            self.connection_status, self.device_config, self.socket_response_processor, self.event_loop
        )
        asyncio.run_coroutine_threadsafe(
            com_backend.create_connection(self.device_config.device_ip, self.device_config.device_port), self.event_loop
        )
        return com_backend

    # Externally accessible disconnect method
    def disconnect(self):
        asyncio.run_coroutine_threadsafe(self.com_backend.destroy_connection(), self.event_loop)

    @abc.abstractmethod
    def send_heartbeat(self, timeout):
        """
        Driver instances must override this method.
        For blocking heartbeat calls, call self.watchdog.feed_watchdog() on success and respect the timeout.
        """
        raise NotImplementedError

    # TODO: implement the following functions later
    # @abc.abstractmethod
    # def sync_state(self):
    #     """
    #     Function to synchronize state before command.
    #     """
    #     raise NotImplementedError

    # @abc.abstractmethod
    # def check_idle(self):
    #     """
    #     Decorator to check whether the state is idle before command.
    #     """
    #     raise NotImplementedError

    # def socket_response_processor(self, message):
    #     """
    #     Driver instances that use socket communication backend must override this method.
    #     For nonblocking heartbeat calls, call self.watchdog.feed_watchdog() here.
    #     """
    #     raise NotImplementedError



"""Communication base for devices"""
import asyncio
import contextlib
from datetime import datetime
import logging
import threading

from shifu_driver.schema import ConnectionState


class SocketBackend:
    def __init__(self, connection_status, device_config, message_processor, event_loop):
        # Start logging
        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(logging.DEBUG)

        # Set config and create connection state attributes
        self.config = device_config
        self.connection = connection_status

        # Set message processor for callback upon response available
        self.message_processor = message_processor

        # Set asyncio event loop
        self.event_loop = event_loop

        # Create connection-dependent variables
        self.send_queue = None
        self.response_available = threading.Event()
        self.device_tasks = set()

        # self.watchdog = watchdog.Watchdog(self.config.heartbeat_send_delay, self, None)

    async def read_handler(self, reader):
        self.logger.debug("read handler started")

        # Wait for message from device and process it
        async for message in self.read(reader):
            await asyncio.to_thread(self.message_processor, message)
            self.response_available.set()

        self.logger.debug("exiting read handler")

    async def read(self, reader):
        data_stream = b''
        try:
            while data_stream := data_stream + await reader.read(1024):
                if self.config.term_char.encode() in data_stream:
                    message, data_stream = data_stream.split(self.config.term_char.encode(), 1)
                    self.connection.packets_received += 1
                    yield message.decode()  # TODO: Handle decode error
        except ConnectionResetError:
            self.logger.warning("Connection reset")
            pass  # TODO: Handle reconnect strategy
        if data_stream:
            yield data_stream.decode()  # TODO: Handle decode error

    async def write_handler(self, writer):
        self.logger.debug("Write handler started")

        try:
            while (message := await self.send_queue.get()) != b'':
                await self.write(writer, message)
                self.send_queue.task_done()
        finally:
            self.logger.debug("Draining writer...")
            await writer.drain()
            writer.close()
            self.logger.debug("Exiting write handler")
            self.logger.info(
                "Disconnected from device uri {self.config.device_ip}:{self.config.device_port}"
            )  # noqa: E501

    async def write(self, writer, message):
        # Terminate message
        if not message.endswith(self.config.term_char):
            message += self.config.term_char

        # Write message
        writer.write(message.encode())
        await writer.drain()  # TODO: Handle broken pipe / other socket errors
        self.connection.packets_sent += 1

    async def add_to_send_queue(self, message):
        if self.send_queue:
            await self.send_queue.put(message)
        else:
            self.logger.warning(f"Send failed, device not connected: {message}")

    async def create_connection(self, host='127.0.0.1', port=9000):
        # Check if device is already connected or a connection attempt is pending
        if self.connection.state == ConnectionState.CONNECTED:
            self.logger.warning("Device already connected")
            return
        elif self.connection.state == ConnectionState.CONNECTION_IN_PROGRESS:
            self.logger.warning("Already attempting to connect to device")
            return

        # Signal that a connection attempt is in progress
        self.connection.state = ConnectionState.CONNECTION_IN_PROGRESS
        print('CONNECTION_IN_PROGRESS!!!')

        # Create asyncio queues for sending and receiving data
        self.send_queue = asyncio.Queue()  # TODO: Convert to PriorityQueue to service ESTOP first

        # Handle server configuration
        if self.config.backend_type == 'server':  # TODO: Implement server handling
            self.logger.info(f"Creating connection on {host}:{port} for device")
            self.server = await asyncio.start_server(self.set_stream_refs, host, port)
        # Handle client configuration
        else:
            self.logger.info(f"Connecting to device on {host}:{port}")
            retry_num = 0
            while retry_num < self.config.max_conn_attempts:
                try:
                    reader, writer = await asyncio.open_connection(host, port)
                    break
                except ConnectionRefusedError:
                    self.logger.warning("Failed to connect to device, retrying...")
                    retry_num += 1
                    await asyncio.sleep(2)
            # Max attempts hit, connection failed
            if retry_num >= self.config.max_conn_attempts:
                self.connection.state = ConnectionState.CONNECTION_FAILED
                self.logger.warning(f"Failed to connect to device on {host}:{port}")
                return

            # Connection was successful
            self.connection.state = ConnectionState.CONNECTED
            self.connection.start_time = datetime.now()
            self.logger.info(f"Connected to {host}:{port}")

        # Create tasks for reading from and writing to device
        read_task = asyncio.create_task(self.read_handler(reader), name='read_handler')
        write_task = asyncio.create_task(self.write_handler(writer), name='write_handler')

        # Add newly created tasks to set of device tasks
        self.device_tasks.add(read_task)
        self.device_tasks.add(write_task)

        # # Start device watchdog
        # self.watchdog.start()

    def set_stream_refs(self, reader, writer):
        self.reader = reader
        self.writer = writer

    async def destroy_connection(self):
        # Ensure device is actually connected
        if self.connection.state != ConnectionState.CONNECTED:
            self.logger.warning("Device not connected")
            return

        self.logger.info(
            f"Disconnecting from device at {self.config.device_ip}:{self.config.device_port}"
        )  # noqa: E501

        # # Stop watchdog
        # self.watchdog.stop()

        # Cancel pending device tasks
        pending_tasks = self.device_tasks.intersection(asyncio.all_tasks())

        for task in pending_tasks:
            task.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await task

        # Destroy queue and clear connection-associated tasks
        self.send_queue = None
        self.response_available.clear()
        self.device_tasks.clear()

        # Reset per-session connection stats
        self.connection.state = ConnectionState.DISCONNECTED
        self.connection.start_time = 0

    # Send data to device
    def send(self, message, wait_for_response=False):
        asyncio.run_coroutine_threadsafe(self.add_to_send_queue(message), self.event_loop)

        # Optionally block until a response is received
        if wait_for_response:
            return self.wait_for_response()
        return True

    def wait_for_response(self):
        # TODO: This is a potential race condition. If the device responds before this function is called,
        #   the flag with be cleared prematurely and will block until the next response or timeout. This
        #   is unlikely to happen but should be noted.
        self.response_available.clear()
        if not self.response_available.wait(self.config.response_timeout):
            self.logger.warning("Response timeout")
            return False
        self.response_available.clear()
        return True



import dataclasses
from dataclasses import dataclass
from datetime import datetime
from enum import Enum, auto
import json


class ExpandedJSONEncoder(json.JSONEncoder):
    """Custom JSON encoder class for handling serialization of additional object types"""

    # Override the default 'default' method to serialize other types
    def default(self, obj):
        if dataclasses.is_dataclass(obj):
            return dataclasses.asdict(obj)
        elif isinstance(obj, datetime):
            return obj.isoformat()
        elif isinstance(obj, Enum):
            return obj.name

        # Call the base class to handle other type or raise TypeError
        return super().default(obj)


# Enum for all possible connection states
class ConnectionState(Enum):
    CONNECTED = auto()
    CONNECTION_IN_PROGRESS = auto()
    CONNECTION_FAILED = auto()
    WATCHDOG_FAILED = auto()
    DISCONNECTED = auto()


# Enum for all possible internal device states
class DeviceState(Enum):
    IDLE = auto()
    IN_USE = auto()
    EMERGENCY_STOPPED = auto()
    ERROR = auto()
    DISABLED = auto()


class JSONObject:
    def to_dict(self):
        return dataclasses.asdict(self)

    def __str__(self):
        return json.dumps(self, indent=4, cls=ExpandedJSONEncoder)


@dataclass
class DeviceStatus(JSONObject):
    # Data class storing info about the current internal state of a device
    state: DeviceState = DeviceState.IDLE
    error_info: str = ""


@dataclass
class ConnectionStatus(JSONObject):
    # Current connection state of device
    state: ConnectionState = ConnectionState.DISCONNECTED
    start_time: datetime = datetime.min
    requests_received: int = 0
    events_published: int = 0
    packets_sent: int = 0
    packets_received: int = 0
    heartbeat_packets_sent: int = 0
    heartbeat_packets_received: int = 0
    heartbeat_failure_count: int = 0
    last_heartbeat_sent: datetime = datetime.min
    last_heartbeat_received: datetime = datetime.min


# Data class representing the configuration of a device
@dataclass
class Config(JSONObject):
    # Connection settings
    device_ip: str = '127.0.0.1'
    device_port: int = 1234
    backend_type: str = 'client'
    term_char: str = '\n'

    # Watchdog settings
    heartbeat_send_delay: float = 1.0  # Rate for sending heartbeat requests (in seconds)
    heartbeat_timeout: int = 2  # Time before device times out (in seconds)
    heartbeat_failure_limit: int = 3  # Number of times a device can time out before being marked dead
    heartbeat_logging: bool = False  # Sets whether to log heartbeat requests and responses

    @staticmethod
    def from_json(json):
        return Config(
            device_ip=json.get('device_ip', '127.0.0.1'),
            device_port=json.get('device_port', '1234'),
            backend_type=json.get('backend_type', 'client'),
            term_char=json.get('term_char', '\n'),
            heartbeat_send_delay=json.get('heartbeat_send_delay', 1.0),
            heartbeat_timeout=json.get('heartbeat_timeout', 2),
            heartbeat_failure_limit=json.get('heartbeat_failure_limit', 3),
            heartbeat_logging=json.get('heartbeat_logging', False),
        )
